# coding: utf-8
from __future__ import unicode_literals

import os
from io import StringIO

from .compat import text
from .exceptions import InvalidBatch
from .utils import guess_mimetype

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, BinaryIO, Dict, List, Optional, Text, Union  # noqa
        from io import FileIO  # noqa
        from .comments import API as CommentsAPI  # noqa
        from .directories import API as DirectoriesAPI  # noqa
        from .documents import API as DocumentsAPI  # noqa
        from .endpoint import APIEndpoint  # noqa
        from .groups import API as GroupsAPI  # noqa
        from .operations import API as OperationsAPI  # noqa
        from .tasks import API as TasksAPI  # noqa
        from .uploads import API as UploadsAPI  # noqa
        from .users import API as UsersAPI  # noqa
        from .workflows import API as WorkflowsAPI  # noqa
except ImportError:
    pass


# Base classes


class Model(object):
    """ Base class for all entities. """

    _valid_properties = {}  # type: Dict[Text, Any]
    service = None  # type: APIEndpoint
    uid = None  # type: Text

    def __init__(self, service=None, **kwargs):
        # type: (Optional[APIEndpoint], Any) -> None
        self.service = service

        # Declare attributes
        for key, default in type(self)._valid_properties.items():
            key = key.replace("-", "_")
            setattr(self, key, kwargs.get(key, default))

    def __repr__(self):
        # type: () -> Text
        attrs = ", ".join(
            "{}={!r}".format(attr, getattr(self, attr.replace("-", "_"), None))
            for attr in sorted(self._valid_properties)
        )
        return "<{} {}>".format(self.__class__.__name__, attrs)

    def as_dict(self):
        # type: () -> Dict[Text, Any]
        """ Returns a dict representation of the resource. """
        result = {}
        for key in self._valid_properties:
            val = getattr(self, key.replace("-", "_"))
            if val is None:
                continue
            # Parse lists of objects
            elif isinstance(val, list) and len(val) > 0 and isinstance(val[0], Model):
                val = [item.as_dict() for item in val]

            result[key] = val
        return result

    @classmethod
    def parse(cls, json, service=None):
        # type: (Dict[Text, Any], Optional[APIEndpoint]) -> Model
        """ Parse a JSON object into a model instance. """
        kwargs = {k: v for k, v in json.items()}
        return cls(service=service, **kwargs)

    def save(self):
        # type: () -> None
        """ Save the resource. """
        self.service.put(self)


class RefreshableModel(Model):
    def load(self, model=None):
        # type: (Optional[Union[Model, Dict[Text, Any]]]) -> None
        """
        Reload the Model.

        If model is not None, copy from its attributes,
        otherwise query the server for the entity with its uid.

        :param model: the entity to copy
        :return: the refreshed model
        """
        if not model:
            model = self.service.get(self.uid)

        if isinstance(model, dict):

            def get_prop(obj, key):
                return obj.get(key)

        else:

            def get_prop(obj, key):
                return getattr(obj, key)

        for key in self._valid_properties:
            key = key.replace("-", "_")
            setattr(self, key, get_prop(model, key))


# Entities


class Batch(Model):
    """ Upload batch. """

    _valid_properties = {
        "batchId": None,
        "blobs": {},
        "dropped": None,
        "extraInfo": None,
        "etag": None,
        "multiPartUploadId": None,
        "provider": None,
        "upload_idx": 0,
    }
    service = None  # type: UploadsAPI

    @property
    def uid(self):
        # type: () -> Text
        return self.batchId

    @uid.setter
    def uid(self, value):
        # type: (Text) -> None
        self.batchId = value

    def cancel(self):
        # type: () -> None
        """ Cancel an upload batch. """
        if not self.batchId:
            return
        self.service.delete(self.uid)
        self.batchId = None

    def delete(self, file_idx):
        """ Delete a blob from the batch. """
        if self.batchId:
            self.service.delete(self.uid, file_idx=file_idx)
            self.blobs[file_idx] = None

    def get(self, file_idx):
        # type: (int) -> Blob
        """
        Get the blob info.

        :param file_idx: the index of the blob in the batch
        :return: the corresponding blob
        """
        if self.batchId is None:
            raise InvalidBatch("Cannot fetch blob for inexistant/deleted batch.")
        blob = self.service.get(self.uid, file_idx=file_idx)
        self.blobs[file_idx] = blob
        return blob

    def get_uploader(self, blob, **kwargs):
        # type: (Blob, Any) -> Blob
        """
        Get an uploader for blob.

        :param blob: the blob to upload
        :param kwargs: the upload settings
        :return: the uploader
        """
        return self.service.get_uploader(self, blob, **kwargs)

    def upload(self, blob, **kwargs):
        # type: (Blob, Any) -> Blob
        """
        Upload a blob.

        :param blob: the blob to upload
        :param kwargs: the upload settings
        :return: the blob info
        """
        return self.service.upload(self, blob, **kwargs)

    def execute(self, operation, file_idx=None, params=None):
        # type: (Text, int, Dict[Text, Any]) -> Any
        """
        Execute an operation on this batch.

        :param operation: operation to execute
        :param file_idx: target file of the operation
        :param params: parameters of the operation
        :return: the output of the operation
        """
        return self.service.execute(self, operation, file_idx, params)

    def attach(self, doc, file_idx=None):
        # type: (Text, Optional[int]) -> Any
        """
        Attach one or all files of this batch to a document.

        :param doc: document to attach
        :param file_idx: target file
        :return: the output of the attach operation
        """
        return self.service.attach(self, doc, file_idx)

    def complete(self):
        # type: () -> Any
        """
        Complete a S3 Direct Upload.

        :return: the output of the complete operation
        """
        return self.service.complete(self)


class Blob(Model):
    """ Blob superclass used for metadata. """

    _valid_properties = {
        "chunkCount": 0,
        "fileIdx": None,
        "mimetype": None,
        "name": None,
        "size": 0,
        "uploadType": None,
        "uploaded": "true",
        "uploadedChunkIds": [],
        "uploadedSize": 0,
    }
    service = None  # type: UploadsAPI

    @classmethod
    def parse(cls, json, service=None):
        # type: (Dict[Text, Any], Optional[APIEndpoint]) -> Blob
        """ Parse a JSON object into a blob instance. """
        kwargs = {k: v for k, v in json.items()}
        model = cls(service=service, **kwargs)

        if model.uploaded and model.uploadedSize == 0:
            model.uploadedSize = model.size
        return model

    def to_json(self):
        # type: () -> Dict[Text, Text]
        """ Return a JSON object used during the upload. """
        return {"upload-batch": self.batch_id, "upload-fileId": text(self.fileIdx)}


class BufferBlob(Blob):
    """
    InMemory content to upload to Nuxeo.

    Acts as a context manager so its data can be read
    with the `with` statement.
    """

    stringio = None  # type: Optional[StringIO]

    def __init__(self, data, **kwargs):
        # type: (Text, Any) -> None
        """
        :param data: content to upload to Nuxeo
        :param **kwargs: named attributes
        """
        super(BufferBlob, self).__init__(**kwargs)
        self.buffer = data
        self.size = len(self.buffer)
        self.mimetype = "application/octet-stream"

    @property
    def data(self):
        # type: () -> StringIO
        """ Request data. """
        return self.stringio

    def __enter__(self):
        if not self.buffer:
            return None
        self.stringio = StringIO(self.buffer)
        return self.stringio

    def __exit__(self, *args):
        if self.stringio:
            self.stringio.close()


class Comment(Model):
    """ Comment. """

    _valid_properties = {
        "ancestorIds": [],
        "author": None,
        "creationDate": None,
        "entity": None,
        "entity-type": "comment",
        "entityId": None,
        "id": None,
        "lastReplyDate": None,
        "modificationDate": None,
        "numberOfReplies": 0,
        "origin": None,
        "parentId": None,
        "text": None,
    }
    service = None  # type: CommentsAPI

    @property
    def uid(self):
        # type: () -> Text
        return self.id

    def delete(self):
        # type: () -> None
        """ Delete the comment. """
        self.service.delete(self.uid)

    def has_replies(self):
        # type: () -> bool
        """ Return True is the comment has at least one reply. """
        return self.numberOfReplies > 0

    def replies(self, **params):
        # type: (Text, Any) -> List[Comment]
        """
        Get the replies of the comment.

        Any additionnal arguments will be passed to the *params* parent's call.

        :param uid: the ID of the comment
        :return: the list of replies
        """
        return self.service.replies(self.uid, params=params)

    def reply(self, text):
        # type: (Text) -> Comment
        """ Add a reply to the comment. """
        # Add the reply
        reply_comment = self.service.post(Comment(parentId=self.uid, text=text))

        # Update comment attributes
        self.numberOfReplies += 1
        self.lastReplyDate = reply_comment.creationDate

        # And return the reply
        return reply_comment


class FileBlob(Blob):
    """
    File to upload to Nuxeo.

    Acts as a context manager so its data can be read
    with the `with` statement.
    """

    # File descriptor
    fd = None  # type: Optional[BinaryIO]

    def __init__(self, path, **kwargs):
        # type: (Text, Any) -> None
        """
        :param path: file path
        :param **kwargs: named attributes
        """
        super(FileBlob, self).__init__(**kwargs)
        self.path = path
        self.name = os.path.basename(self.path)
        self.size = os.path.getsize(self.path)
        self.mimetype = self.mimetype or guess_mimetype(self.path)  # type: Text

    @property
    def data(self):
        # type: () -> BinaryIO
        """
        Request data.

        The caller has to close the file descriptor
        himself if he doesn't open it with the
        context manager.
        """
        return self.fd

    def __enter__(self):
        self.fd = open(self.path, "rb")
        return self.fd

    def __exit__(self, *args):
        if self.fd:
            self.fd.close()


class Directory(Model):
    """ Directory. """

    _valid_properties = {
        "directoryName": None,
        "entries": [],
        "entity-type": "directory",
    }
    service = None  # type: DirectoriesAPI

    @property
    def uid(self):
        # type: () -> Text
        return self.directoryName

    def get(self, entry):
        # type: (Text) -> DirectoryEntry
        """
        Get an entry of the directory.

        :param entry: the name of the entry
        :return: the corresponding directory entry
        """
        return self.service.get(self.uid, dir_entry=entry)

    def create(self, entry):
        # type: (DirectoryEntry) -> DirectoryEntry
        """ Create an entry in the directory. """
        return self.service.post(entry, dir_name=self.uid)

    def save(self, entry):
        # type: (DirectoryEntry) -> DirectoryEntry
        """ Save a modified entry of the directory. """
        return self.service.put(entry, dir_name=self.uid)

    def delete(self, entry):
        # type: (Text) -> None
        """
        Delete one of the directory's entries.

        :param entry: the entry to delete
        """
        self.service.delete(self.uid, entry)

    def exists(self, entry):
        # type: (Text) -> bool
        """
        Check if an entry is in the directory.

        :param entry: the entry name
        :return: True if it exists, else False
        """
        return self.service.exists(self.uid, dir_entry=entry)


class DirectoryEntry(Model):
    """ Directory entry. """

    _valid_properties = {
        "directoryName": None,
        "entity-type": "directoryEntry",
        "properties": {},
    }
    service = None  # type: DirectoriesAPI

    @property
    def uid(self):
        # type: () -> Text
        return self.properties["id"]

    def save(self):
        # type: () -> DirectoryEntry
        """ Save the entry. """
        return self.service.put(self, self.directoryName)

    def delete(self):
        # type: () -> None
        """ Delete the entry. """
        self.service.delete(self.directoryName, self.uid)


class Document(RefreshableModel):
    """ Document. """

    _valid_properties = {
        "changeToken": None,
        "contextParameters": {},
        "entity-type": "document",
        "facets": [],
        "isCheckedOut": False,
        "isProxy": False,
        "isTrashed": False,
        "isVersion": False,
        "lastModified": None,
        "name": None,
        "parentRef": None,
        "path": None,
        "properties": {},
        "repository": "default",
        "state": None,
        "title": None,
        "type": None,
        "uid": None,
        "versionLabel": None,
    }
    service = None  # type: DocumentsAPI

    @property
    def workflows(self):
        # type: () -> List[Workflow]
        """ Return the workflows associated with the document. """
        return self.service.workflows(self)

    def add_permission(self, params):
        # type: (Dict[Text, Any]) -> None
        """
        Add a permission to a document.

        :param params: permission to add
        """
        return self.service.add_permission(self.uid, params)

    def comment(self, text):
        # type: (Text) -> Comment
        """
        Add a comment to the document.

        :param text: the comment message
        :return: a comment object
        """
        return self.service.comment(self.uid, text)

    def comments(self, **params):
        # type: (Any) -> List[Comment]
        """Return the comments associated with the document.
        Any additionnal arguments will be passed to the *params* parent's call.
        """
        return self.service.comments(self.uid, params=params)

    def convert(self, params):
        # type: (Dict[Text, Any]) -> Union[Dict[Text, Any], Text]
        """
        Convert the document to another format.

        :param params: Converter permission
        :return: the converter result
        """
        return self.service.convert(self.uid, params)

    def delete(self):
        # type: () -> None
        """ Delete the document. """
        self.service.delete(self.uid)

    def fetch_acls(self):
        # type: () -> Dict[Text, Any]
        """ Fetch document ACLs. """
        return self.service.fetch_acls(self.uid)

    def fetch_audit(self):
        # type: () -> Dict[Text, Any]
        """ Fetch audit for current document. """
        return self.service.fetch_audit(self.uid)

    def fetch_blob(self, xpath="blobholder:0"):
        # type: (Text) -> Blob
        """
        Retrieve one of the blobs attached to the document.

        :param xpath: the xpath to the blob
        :return: the blob
        """
        return self.service.fetch_blob(uid=self.uid, xpath=xpath)

    def fetch_lock_status(self):
        # type: () -> Dict[Text, Any]
        """ Get lock informations. """
        return self.service.fetch_lock_status(self.uid)

    def fetch_rendition(self, name):
        # type: (Text) -> Union[Text, bytes]
        """
        :param name: Rendition name to use
        :return: The rendition content
        """
        return self.service.fetch_rendition(self.uid, name)

    def fetch_renditions(self):
        # type: () -> List[Union[Text, bytes]]
        """
        :return: Available renditions for this document
        """
        return self.service.fetch_renditions(self.uid)

    def follow_transition(self, name):
        # type: (Text) -> None
        """
        Follow a lifecycle transition on this document.

        :param name: transition name
        """
        doc = self.service.follow_transition(self.uid, name)
        self.load(doc)

    def get(self, prop):
        # type: (Text) -> Any
        """ Get a property of the document by its name. """
        return self.properties[prop]

    def has_permission(self, permission):
        # type: (Text) -> bool
        """ Verify if a document has the permission. """
        return self.service.has_permission(self.uid, permission)

    def is_locked(self):
        # type: () -> bool
        """ Get the lock status. """
        return not not self.fetch_lock_status()

    def lock(self):
        # type: () -> Dict[Text, Any]
        """ Lock the document. """
        return self.service.lock(self.uid)

    def move(self, dst, name=None):
        # type: (Text, Optional[Text]) -> None
        """
        Move a document into another parent.

        :param dst: The new parent path
        :param name: Rename the document if specified
        """
        doc = self.service.move(self.uid, dst, name)
        self.load(doc)

    def remove_permission(self, params):
        # type: (Dict[Text, Any]) -> None
        """ Remove a permission to a document. """
        return self.service.remove_permission(self.uid, params)

    def set(self, properties):
        # type: (Dict[Text, Any]) -> None
        """ Add/update the properties of the document. """
        self.properties.update(properties)

    def trash(self):
        # type: () -> None
        doc = self.service.trash(self.uid)
        self.load(doc)

    def unlock(self):
        # type: () -> Dict[Text, Any]
        """ Unlock the document. """
        return self.service.unlock(self.uid)

    def untrash(self):
        # type: () -> None
        doc = self.service.untrash(self.uid)
        self.load(doc)


class Group(Model):
    """ User group. """

    _valid_properties = {
        "entity-type": "group",
        "grouplabel": None,
        "groupname": None,
        "memberGroups": [],
        "memberUsers": [],
    }
    service = None  # type: GroupsAPI

    @property
    def uid(self):
        # type: () -> Text
        return self.groupname

    def delete(self):
        # type: () -> None
        """ Delete the group. """
        self.service.delete(self.uid)


class Operation(Model):
    """ Automation operation. """

    _valid_properties = {
        "command": None,
        "context": None,
        "input_obj": None,
        "params": {},
        "progress": 0,
    }
    service = None  # type: OperationsAPI

    def execute(self, **kwargs):
        # type: (Any) -> Any
        """ Execute the operation. """
        return self.service.execute(self, **kwargs)


class Task(RefreshableModel):
    """ Workflow task. """

    _valid_properties = {
        "actors": [],
        "comments": [],
        "created": None,
        "directive": None,
        "dueDate": None,
        "entity-type": "task",
        "id": None,
        "name": None,
        "nodeName": None,
        "state": None,
        "targetDocumentIds": [],
        "taskInfo": {},
        "variables": {},
        "workflowInstanceId": None,
        "workflowModelName": None,
    }
    service = None  # type: TasksAPI

    @property
    def uid(self):
        # type: () -> Text
        return self.id

    def complete(self, action, variables=None, comment=None):
        # type: (Text, Optional[Dict[Text, Any]], Optional[Text]) -> None
        """ Complete the action of a task. """
        updated_task = self.service.complete(
            self, action, variables=variables, comment=comment
        )
        self.load(updated_task)

    def delegate(self, actors, comment=None):
        # type: (Text, Optional[Text]) -> None
        """ Delegate the task to someone else. """
        self.service.transfer(self, "delegate", actors, comment=comment)
        self.load()

    def reassign(self, actors, comment=None):
        # type: (Text, Optional[Text]) -> None
        """ Reassign the task to someone else. """
        self.service.transfer(self, "reassign", actors, comment=comment)
        self.load()


class User(RefreshableModel):
    """ User. """

    _valid_properties = {
        "entity-type": "user",
        "extendedGroups": [],
        "id": None,
        "isAdministrator": False,
        "isAnonymous": False,
        "properties": {},
    }
    service = None  # type: UsersAPI

    @property
    def uid(self):
        # type: () -> Text
        return self.id

    def change_password(self, password):
        # type: (Text) -> None
        """
        Change user password.

        :param password: New password to set
        """
        self.properties["password"] = password
        self.save()

    def delete(self):
        # type: () -> None
        """ Delete the user. """
        self.service.delete(self.uid)


class Workflow(Model):
    """ Workflow. """

    _valid_properties = {
        "attachedDocumentIds": [],
        "entity-type": "workflow",
        "graphResource": None,
        "id": None,
        "initiator": None,
        "name": None,
        "state": None,
        "title": None,
        "variables": {},
        "workflowModelName": None,
    }
    service = None  # type: WorkflowsAPI

    @property
    def tasks(self):
        # type: () -> List[Task]
        """ Return the tasks associated with the workflow. """
        return self.service.tasks(self)

    @property
    def uid(self):
        # type: () -> Text
        return self.id

    def delete(self):
        # type: () -> None
        """ Delete the workflow. """
        self.service.delete(self.uid)

    def graph(self):
        # type: () -> Dict[Text, Any]
        """ Return a JSON representation of the workflow graph. """
        return self.service.graph(self)



# coding: utf-8
from __future__ import unicode_literals

from .compat import text
from .constants import UPLOAD_CHUNK_SIZE, UP_AMAZON_S3
from .endpoint import APIEndpoint
from .exceptions import HTTPError, InvalidUploadHandler, UploadError
from .models import Batch, Blob
from .utils import SwapAttr, chunk_partition

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import (
            Any,
            BinaryIO,
            Callable,
            Dict,
            List,
            Optional,
            Text,
            Tuple,
            Union,
        )  # noqa
        from .client import NuxeoClient  # noqa
        from .models import BufferBlob, FileBlob  # noqa

        ActualBlob = Union[BufferBlob, FileBlob]  # noqa
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for uploads. """

    def __init__(self, client, endpoint="upload", headers=None):
        # type: (NuxeoClient, Text, Optional[Dict[Text, Text]]) -> None
        super(API, self).__init__(client, endpoint=endpoint, cls=Blob, headers=headers)

        # Available upload handlers
        self.__handlers = None  # type: Optional[List[str]]

    def get(self, batch_id, file_idx=None):
        # type: (Text, Optional[int]) -> Union[List[Blob], Blob]
        """
        Get the detail of a batch.

        If file_idx is None, returns the details of all its blobs,
        otherwise returns the details of the corresponding blob.

        :param batch_id: the id of the batch
        :param file_idx: the index of the blob
        :return: the batch details
        """
        path = batch_id
        if file_idx is not None:
            path = "{}/{}".format(path, file_idx)

        resource = super(API, self).get(path=path)

        if file_idx is not None:
            resource.batch_id = batch_id
            resource.fileIdx = file_idx
        elif not resource:
            return []
        return resource

    def post(self, handler=""):
        # type: (Optional[Text]) -> Batch
        """
        Create a batch.

        :param handler: the upload handler to use
        :return: the created batch
        """
        endpoint = self.endpoint
        if handler:
            handler = handler.lower()
            handlers = self.handlers()
            if handler in handlers:
                if handler != "default":
                    endpoint = "{}/new/{}".format(endpoint, handler)
            else:
                raise InvalidUploadHandler(handler, handlers)

        response = self.client.request("POST", endpoint)
        return Batch.parse(response.json(), service=self)

    batch = post  # Alias for clarity

    def put(self, **kwargs):
        # type: (Any) -> None
        raise NotImplementedError()

    def delete(self, batch_id, file_idx=None):
        # type: (Text, Optional[int]) -> None
        """
        Delete a batch or a blob.

        If the file_idx is None, deletes the batch,
        otherwise deletes the corresponding blob.

        :param batch_id: the id of the batch
        :param file_idx: the index of the blob
        """
        if file_idx is not None:
            target = "{}/{}".format(batch_id, file_idx)
            super(API, self).delete(target)
        else:
            target = batch_id
            with SwapAttr(self, "_cls", Batch):
                super(API, self).delete(target)

    def handlers(self, force=False):
        # type: (Optional[bool]) -> List[str]
        """
        Get available upload handlers.

        :param force: force refreshing the list
        """
        if self.__handlers is None or force:
            endpoint = "{}/handlers".format(self.endpoint)
            try:
                response = self.client.request("GET", endpoint)
                self.__handlers = [h for h in response.json()["handlers"][0].values()]
            except Exception:
                # This is not good, no handlers == no uploads!
                # Return an empty list without modifying .__handlers
                # to force a new HTTP call the next time.
                return []
        return self.__handlers

    def send_data(
        self,
        name,  # type: Text
        data,  # type: Union[BinaryIO, Text, bytes]
        path,  # type: Text
        chunked,  # type: bool
        index,  # type: int
        headers,  # type: Dict[Text, Text]
        data_len=0,  # type: Optional[int]
    ):
        # type: (...) -> Blob
        """
        Send data/chunks to the server.

        :param name: name of the file being uploaded
        :param data: data being sent
        :param path: url for the upload
        :param chunked: True if the upload is in chunks
        :param index: which chunk is being sent (0 if not chunked)
        :param headers: HTTP request headers
        :return: the blob info
        """
        if chunked:
            headers["X-Upload-Chunk-Index"] = text(index)

        if data_len > 0:
            headers["Content-Length"] = text(data_len)

        try:
            return super(API, self).post(
                resource=data, path=path, raw=True, headers=headers
            )
        except HTTPError as e:
            raise UploadError(name, chunk=index if chunked else None, info=str(e))

    def state(self, path, blob, chunk_size=UPLOAD_CHUNK_SIZE):
        # type: (Text, ActualBlob, int) -> Tuple[int, List[int]]
        """
        Get the state of a blob.

        If the blob upload has not begun yet, the server
        will return a 404 error, so we initialize the
        different values.
        If the blob upload is incomplete, we return the
        values the server sent us.
        If the blob upload is complete, we return None
        for these values.

        :param path: path for the request
        :param blob: the target blob
        :param chunk_size: the chunk size for new uploads
        :return: a tuple of the chunk count and
                 the set of uploaded chunk indexes
        """
        info = super(API, self).get(path, default=None)

        if info:
            chunk_count = int(info.chunkCount)
            uploaded_chunks = [int(i) for i in info.uploadedChunkIds]
        else:
            # It's a new upload
            chunk_count, _ = chunk_partition(blob.size, chunk_size)
            uploaded_chunks = []

        return chunk_count, uploaded_chunks

    def upload(
        self,
        batch,  # type: Batch
        blob,  # type: ActualBlob
        chunked=False,  # type: bool
        chunk_size=UPLOAD_CHUNK_SIZE,  # type: int
        callback=None,  # type: Union[Callable, Tuple[Callable]]
    ):
        # type: (...) -> Blob
        """
        Upload a blob.

        Can be used to upload a new blob or resume
        the upload of a chunked blob.

        :param batch: batch of the upload
        :param blob: blob to upload
        :param chunked: if True, send in chunks
        :param chunk_size: if blob is bigger, send in chunks of this size
        :param callback: if not None, they are executed between each chunk.
          It is either a single callable or a tuple of callables (tuple is used to keep order).
        :return: uploaded blob details
        """
        uploader = self.get_uploader(
            batch, blob, chunked, chunk_size, callback=callback
        )
        uploader.upload()
        return uploader.blob

    def execute(self, batch, operation, file_idx=None, params=None, void_op=True):
        # type: (Batch, Text, Optional[int], Optional[Dict[Text,Any]], bool) -> Any
        """
        Execute an operation with the batch or one of its files as an input.

        :param batch: input for the operation
        :param operation: operation to execute
        :param file_idx: if not None, sole input of the operation
        :param params: parameters for the operation
        :param void_op: if True, the body of the response
        from the server will be empty
        :return: the output of the operation
        """
        path = "{}/{}".format(self.endpoint, batch.uid)
        if file_idx is not None:
            path = "{}/{}".format(path, file_idx)

        path = "{}/execute/{}".format(path, operation)

        headers = {}
        if void_op:
            headers = {"X-NXVoidOperation": "true"}

        return self.client.request(
            "POST", path, data={"params": params}, headers=headers or None
        )

    def attach(self, batch, doc, file_idx=None):
        # type: (Batch, Text, Optional[int]) -> Any
        """
        Attach one or all files of a batch to a document.

        :param batch: batch to attach
        :param doc: document to attach
        :param file_idx: if not None, only this file will be attached
        :return: the output of the attach operation
        """
        params = {"document": doc}
        if file_idx is None and batch.upload_idx > 1:
            params["xpath"] = "files:files"
        return self.execute(batch, "Blob.Attach", file_idx, params)

    def complete(self, batch):
        # type: () -> Any
        """
        Complete an upload.
        This is a no-op when using the default upload provider.

        :return: the output of the complete operation
        """
        if batch.provider == UP_AMAZON_S3:
            blob = batch.blobs[0]
            s3_info = batch.extraInfo
            params = {
                "name": blob.name,
                "fileSize": blob.size,
                "key": "{}/{}".format(s3_info["baseKey"].rstrip("/"), blob.name),
                "bucket": s3_info["bucket"],
                "etag": batch.etag,
            }
            endpoint = "{}/{}/{}/complete".format(
                self.endpoint, batch.uid, batch.upload_idx - 1
            )
            return self.client.request("POST", endpoint, data=params)

        # Doing a /complete with the default upload provider
        # will end on a HTTP 409 Conflict error.
        return None

    def get_uploader(
        self,
        batch,  # type: Batch
        blob,  # type: ActualBlob
        chunked=False,  # type: bool
        chunk_size=UPLOAD_CHUNK_SIZE,  # type: int
        callback=None,  # type: Union[Callable, Tuple[Callable]]
    ):
        # type: (...) -> "Uploader"
        """
        Get an upload helper for blob.

        Can be used to upload a new blob or resume
        the upload of a chunked blob.

        :param batch: batch of the upload
        :param blob: blob to upload
        :param chunked: if True, send in chunks
        :param chunk_size: if blob is bigger, send in chunks of this size
        :param callback: if not None, they are executed between each chunk.
          It is either a single callable or a tuple of callables (tuple is used to keep order).
        :return: uploaded blob details
        """
        chunked = chunked and blob.size > chunk_size

        if batch.provider == UP_AMAZON_S3:
            if chunked:
                from .handlers.s3 import ChunkUploaderS3 as cls
            else:
                from .handlers.s3 import UploaderS3 as cls
        else:
            if chunked:
                from .handlers.default import ChunkUploader as cls
            else:
                from .handlers.default import Uploader as cls

        return cls(self, batch, blob, chunk_size, callback)



# coding: utf-8
from __future__ import unicode_literals

from .comments import API as CommentsAPI
from .endpoint import APIEndpoint
from .exceptions import BadQuery, HTTPError, UnavailableConvertor
from .models import Comment, Document
from .utils import SwapAttr
from .workflows import API as WorkflowsAPI

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text, Union  # noqa
        from .client import NuxeoClient  # noqa
        from .models import Blob, Comment, Workflow  # noqa
        from .operations import API as OperationsAPI  # noqa
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for documents. """

    def __init__(
        self,
        client,  # type: NuxeoClient
        operations,  # type: OperationsAPI
        workflows,  # type: WorkflowsAPI
        comments,  # type: CommentsAPI
        endpoint=None,  # type: Text
        headers=None,  # type: Optional[Dict[Text, Text]]
    ):
        # type: (...) -> None
        self.operations = operations
        self.comments_api = comments
        self.workflows_api = workflows
        super(API, self).__init__(
            client, endpoint=endpoint, cls=Document, headers=headers
        )

    def get(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> Document
        """
        Get the detail of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :return: the document
        """
        return super(API, self).get(path=self._path(uid=uid, path=path))

    def post(self, document, parent_id=None, parent_path=None):
        # type: (Document, Optional[Text], Optional[Text]) -> Document
        """
        Create a document.

        :param document: the document to create
        :param parent_id: the id of the parent document
        :param parent_path: the path of the parent document
        :return: the created document
        """
        return super(API, self).post(
            document, path=self._path(uid=parent_id, path=parent_path)
        )

    create = post  # Alias for clarity

    def put(self, document):
        # type: (Document) -> Document
        """
        Update a document.

        :param document: the document to update
        :return: the updated document
        """
        return super(API, self).put(document, path=self._path(uid=document.uid))

    def delete(self, document_id):
        # type: (Text) -> None
        """
        Delete a document.

        :param document_id: the id of the document to delete
        """
        super(API, self).delete(self._path(uid=document_id))

    def exists(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> bool
        """
        Check if a document exists.

        :param uid: the id of the document to check
        :param path: the path of the document to check
        :return: True if it exists, else False
        """
        try:
            self.get(uid=uid, path=path)
            return True
        except HTTPError as e:
            if e.status != 404:
                raise e
        return False

    def add_permission(self, uid, params):
        # type: (Text, Dict[Text, Any]) -> None
        """
        Add a permission to a document.

        :param uid: the uid of the document
        :param params: the permissions to add
        """
        self.operations.execute(
            command="Document.AddPermission", input_obj=uid, params=params
        )

    def comment(self, uid, text):
        # type: (Text, Text) -> Comment
        """
        Add a new comment for a given document.

        :param uid: the ID of the document
        :param text: the content of the comment
        :return: the comment
        """
        obj = Comment(parentId=uid, text=text)
        return self.comments_api.post(obj)

    def comments(self, uid, params=None):
        # type: (Text, Any) -> List[Comment]
        """
        Get the comments of a document.

        Any additionnal arguments will be passed to the *params* parent's call.

        :param uid: the ID of the document
        :return: the list of comments
        """
        path = "id/{}/@comment".format(uid)

        # Adding "&fetch.comment=repliesSummary" to the URL to retrieve replies number as well
        kwargs = {"fetch.comment": "repliesSummary"}
        if isinstance(params, dict):
            kwargs.update(params)

        with SwapAttr(self.comments_api, "endpoint", self.endpoint):
            return super(CommentsAPI, self.comments_api).get(path=path, params=kwargs)

    def convert(self, uid, options):
        # type: (Text, Dict[Text, Text]) -> Union[Text, Dict[Text, Any]]
        """
        Convert a blob into another format.

        :param uid: the uid of the blob to be converted
        :param options: the target type, target format,
                        or converter for the blob
        :return: the response from the server
        """
        xpath = options.pop("xpath", "blobholder:0")
        adapter = "blob/{}/@convert".format(xpath)
        if (
            "converter" not in options
            and "type" not in options
            and "format" not in options
        ):
            raise BadQuery("One of (converter, type, format) is mandatory in options")

        try:
            return super(API, self).get(
                path=self._path(uid=uid), params=options, adapter=adapter, raw=True
            )
        except HTTPError as e:
            if "is not registered" in e.message:
                raise BadQuery(e.message)
            if (
                "is not available" in e.message
                or "UnsupportedOperationException" in e.message
            ):
                raise UnavailableConvertor(options)
            raise e

    def fetch_acls(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the ACLs of a document.

        :param uid: the uid of the document
        :return: the ACLs
        """
        req = super(API, self).get(
            path=self._path(uid=uid), cls=dict, headers=self.headers, enrichers=["acls"]
        )
        return req["contextParameters"]["acls"]

    def fetch_audit(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the audit of a document.

        :param uid: the uid of the document
        :return: the audit
        """
        return super(API, self).get(self._path(uid=uid), adapter="audit", cls=dict)

    def fetch_lock_status(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the lock status of a document.

        :param uid: the uid of the document
        :return: the lock status
        """
        headers = self.headers or {}
        headers.update({"fetch-document": "lock"})
        req = super(API, self).get(path=self._path(uid=uid), cls=dict, headers=headers)
        if "lockOwner" in req:
            return {"lockCreated": req["lockOwner"], "lockOwner": req["lockOwner"]}
        else:
            return {}

    def fetch_rendition(self, uid, name):
        # type: (Text, Text) -> Union[Text, bytes]
        """
        Fetch a rendition of a document.

        :param uid: the uid of the document
        :param name: the name of the rendition
        :return: the corresponding rendition
        """
        adapter = "rendition/{}".format(name)
        return super(API, self).get(path=self._path(uid=uid), raw=True, adapter=adapter)

    def fetch_renditions(self, uid):
        # type: (Text) -> List[Union[Text, bytes]]
        """
        Fetch all renditions of a document.

        :param uid: the uid of a document
        :return: the renditions
        """
        headers = self.headers or {}
        headers.update({"enrichers-document": "renditions"})

        req = super(API, self).get(path=self._path(uid=uid), cls=dict, headers=headers)
        return [rend["name"] for rend in req["contextParameters"]["renditions"]]

    def follow_transition(self, uid, name):
        # type: (Text, Text) -> Dict[Text, Any]
        """
        Follow a lifecycle transition.

        :param uid: the uid of the target document
        :param name: the name of the transition
        """
        params = {"value": name}
        return self.operations.execute(
            command="Document.FollowLifecycleTransition", input_obj=uid, params=params
        )

    def fetch_blob(self, uid=None, path=None, xpath="blobholder:0"):
        # type: (Optional[Text], Optional[Text], Text) -> Blob
        """
        Get the blob of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :param xpath: the xpath of the blob
        :return: the blob
        """
        adapter = "blob/{}".format(xpath)
        return super(API, self).get(
            path=self._path(uid=uid, path=path), raw=True, adapter=adapter
        )

    def get_children(self, uid=None, path=None, enrichers=None):
        # type: (Optional[Text], Optional[Text], Optional[List[str]]) -> List[Document]
        """
        Get the children of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :param enrichers: additionnal details to fetch at the same time, e.g.: ["permissions"]
        :return: the document children
        """
        return super(API, self).get(
            path=self._path(uid=uid, path=path), enrichers=enrichers, adapter="children"
        )

    def has_permission(self, uid, permission):
        # type: (Text, Text) -> bool
        """
        Check if a document has a permission.

        :param uid: the uid of the document
        :param permission: the permission to check
        :return: True if the document has it, False otherwise
        """
        req = super(API, self).get(
            path=self._path(uid=uid),
            cls=dict,
            headers=self.headers,
            enrichers=["permissions"],
        )
        return permission in req["contextParameters"]["permissions"]

    def lock(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """ Lock a document. """
        return self.operations.execute(command="Document.Lock", input_obj=uid)

    def move(self, uid, dst, name=None):
        # type: (Text, Text, Optional[Text]) -> Dict[Text, Any]
        """
        Move a document and eventually rename it.

        :param uid: the uid of the target document
        :param dst: the destination
        :param name: the new name
        """
        params = {"target": dst}
        if name:
            params["name"] = name
        return self.operations.execute(
            command="Document.Move", input_obj=uid, params=params
        )

    def query(self, opts=None):
        # type: (Optional[Dict[Text, Text]]) -> Dict[Text, Any]
        """
        Run a query on the documents.

        :param opts: a query or a pageProvider
        :return: the corresponding documents
        """
        opts = opts or {}
        if "query" in opts:
            query = "NXQL"
        elif "pageProvider" in opts:
            query = opts["pageProvider"]
        else:
            raise BadQuery("Need either a pageProvider or a query")

        path = "query/{}".format(query)
        res = super(API, self).get(path=path, params=opts, cls=dict)
        res["entries"] = [
            Document.parse(entry, service=self) for entry in res["entries"]
        ]
        return res

    def remove_permission(self, uid, params):
        # type: (Text, Dict[Text, Text]) -> None
        """
        Remove a permission on a document.

        :param uid: the uid of the document
        :param params: the permission to remove
        """
        self.operations.execute(
            command="Document.RemovePermission", input_obj=uid, params=params
        )

    def trash(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Trash the document.

        :param uid: the uid of the document
        """
        return self.operations.execute(command="Document.Trash", input_obj=uid)

    def unlock(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """ Unlock a document. """
        return self.operations.execute(command="Document.Unlock", input_obj=uid)

    def untrash(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Untrash the document.

        :param uid: the uid of the document
        """
        return self.operations.execute(command="Document.Untrash", input_obj=uid)

    def workflows(self, document):
        # type: (Document) -> Union[Workflow, List[Workflow]]
        """ Get the workflows of a document. """
        path = "id/{}/@workflow".format(document.uid)

        with SwapAttr(self.workflows_api, "endpoint", self.endpoint):
            return super(WorkflowsAPI, self.workflows_api).get(path=path)

    def _path(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> Text
        if uid:
            path = "repo/{}/id/{}".format(self.client.repository, uid)
        elif path:
            path = "repo/{}/path{}".format(self.client.repository, path)
        return path



# coding: utf-8
from __future__ import unicode_literals

from .compat import text
from .endpoint import APIEndpoint
from .exceptions import BadQuery
from .models import Task

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text, Union  # noqa
        from .client import NuxeoClient  # noqa
        from .models import Workflow  # noqa
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for tasks. """

    def __init__(self, client, endpoint="task", headers=None):
        # type: (NuxeoClient, Text, Optional[Dict[Text, Text]]) -> None
        super(API, self).__init__(client, endpoint=endpoint, cls=Task, headers=headers)

    def get(
        self, options=None  # type: Optional[Union[Dict[Text, Any], Text]]
    ):
        # type: (...) -> Union[Task, List[Task]]
        """
        Get tasks by id or by options.

        :param options: the id of the task or the constraints
        :return: the task(s)
        """
        params, request_path = None, None

        if isinstance(options, dict):
            params = options
        elif isinstance(options, text):
            request_path = options

        return super(API, self).get(path=request_path, params=params)

    def post(self, **kwargs):
        # type: (Any) -> None
        raise NotImplementedError()

    def put(self, task):
        # type: (Task) -> Task
        raise NotImplementedError()

    def delete(self, task_id):
        # type: (Text) -> None
        raise NotImplementedError()

    def complete(self, task, action, variables=None, comment=None):
        # type: (Task, Text, Optional[Dict[Text, Any]], Optional[Text]) -> Task
        """
        Complete the task.

        :param task: the task
        :param action: to take
        :param variables: to add to the Task
        :param comment: for the action
        :return: Updated task
        """
        task.comment = comment
        if variables:
            task.variables.update(variables)

        request_path = "{}/{}".format(task.uid, action)
        return super(API, self).put(task, path=request_path)

    def transfer(self, task, transfer, actors, comment=None):
        # type: (Task, Text, Text, Optional[Text]) -> None
        """
         Delegate or reassign the Task to someone else.

        :param task: the task to modify
        :param transfer: 'delegate' or 'reassign'
        :param actors: the actors involved
        :param comment: a comment
        :return:
        """
        if transfer == "delegate":
            actors_type = "delegatedActors"
        elif transfer == "reassign":
            actors_type = "actors"
        else:
            raise BadQuery("Task transfer must be either delegate or reassign.")

        params = {actors_type: actors}
        if comment:
            params["comment"] = comment

        request_path = "{}/{}".format(task.uid, transfer)
        super(API, self).put(None, path=request_path, params=params)



# coding: utf-8
from __future__ import unicode_literals

import hashlib
import logging
import mimetypes
import sys

from nuxeo.constants import UP_AMAZON_S3

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from _hashlib import HASH  # noqa
        from typing import Any, Dict, List, Optional, Text, Tuple, Type, Union  # noqa
except ImportError:
    pass


logger = logging.getLogger(__name__)
WIN32_PATCHED_MIME_TYPES = {
    "image/pjpeg": "image/jpeg",
    "image/x-png": "image/png",
    "image/bmp": "image/x-ms-bmp",
    "audio/x-mpg": "audio/mpeg",
    "video/x-mpeg2a": "video/mpeg",
    "application/x-javascript": "application/javascript",
    "application/x-msexcel": "application/vnd.ms-excel",
    "application/x-mspowerpoint": "application/vnd.ms-powerpoint",
    "application/x-mspowerpoint.12": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
}


def chunk_partition(file_size, desired_chunk_size, handler=""):
    # type: (int, int, Optional[Text]) -> Tuple[int, int]
    """Determine the chunk count and chunk size from
    given *file_size* and *desired_chunk_size*.

    There may be boundaries depending of the given upload *handler*.

    :param file_size: the file size to upload (aka the blob)
    :param desired_chunk_size: the desired chunk size
    :param handler: the upload handler to use
    :return: a tuple of the chunk count and chunk size
    """

    chunk_size_min = 1
    chunk_count_max = None

    # Amazon S3 has some limitations
    if handler == UP_AMAZON_S3:
        chunk_size_min = 1024 * 1024 * 5  # 5 MiB
        chunk_count_max = 10000

    # Unsure to have the minimum valid chunk size
    chunk_size = max(chunk_size_min, desired_chunk_size)

    # Compute the number of chunks
    chunk_count = file_size // chunk_size + (file_size % chunk_size > 0)

    # If the chunks count exceeds the maximum chunks count ...
    if chunk_count_max and chunk_count > chunk_count_max:
        # ... In that case, the chunk count will define the chunk size
        return chunk_partition(
            file_size, file_size // (chunk_count_max - 1), handler=handler
        )

    return chunk_count, chunk_size


def get_digest_algorithm(digest):
    # type: (Text) -> Optional[Text]

    # Available algorithms
    digesters = {
        32: "md5",
        40: "sha1",
        56: "sha224",
        64: "sha256",
        96: "sha384",
        128: "sha512",
    }

    # Ensure the digest is hexadecimal
    try:
        int(digest, 16) >= 0
    except (TypeError, ValueError):
        return None

    return digesters.get(len(digest), None)


def get_digest_hash(algorithm):
    # type: (Text) -> Optional[HASH]

    # Retrieve the hashlib function for the given digest, None if not found
    func = getattr(hashlib, algorithm, None)

    return func() if func else None


def get_digester(digest):
    # type: (Text) -> Optional[HASH]
    """
    Get the digester corresponding to the given hash.

    To choose the digester used by the server, see
    https://doc.nuxeo.com/nxdoc/file-storage-configuration/#configuring-the-default-blobprovider

    :param digest: the hash
    :return: the digester function
    """

    algo = get_digest_algorithm(digest)
    func = get_digest_hash(algo) if algo else None

    if not func:
        logger.warning("No valid hash algorithm found for digest %r", digest)

    return func


def guess_mimetype(filename):
    # type: (Text) -> Text
    """ Guess the mimetype of a given file. """
    mime_type, _ = mimetypes.guess_type(filename)
    if mime_type:
        if sys.platform == "win32":
            # Patch bad Windows MIME types
            # See http://bugs.python.org/issue15207 and NXP-11660
            mime_type = WIN32_PATCHED_MIME_TYPES.get(mime_type, mime_type)
        return mime_type

    return "application/octet-stream"


def json_helper(obj):
    # type: (Any) -> Dict[Text, Any]
    return obj.to_json()


def log_chunk_details(chunk_count, chunk_size, uploaded_chunks):
    # type: (int, int, List[int]) -> None
    """Simple helper to log an chunked upload details about chunks data.

    :param chunk_count: the number of chunks
    :param chunk_size: the size of each chunks
    :param uploaded_chunks: the list of already uploaded chunks
    """
    if chunk_count <= 1 or logger.getEffectiveLevel() > logging.DEBUG:
        return

    uploaded_chunks_count = len(uploaded_chunks)
    uploaded_data_length = chunk_size * uploaded_chunks_count
    msg = (
        "Computed chunks count is {:,d}"
        "; chunks size is {:,d} bytes"
        "; uploaded chunks count is {:,d}"
        " => uploaded data so far is {:,d} bytes."
    )
    details = msg.format(
        chunk_count, chunk_size, uploaded_chunks_count, uploaded_data_length
    )
    logger.debug(details)


class SwapAttr(object):
    """
    Context manager to swap an attribute's value:

        >>> # self.person equals 'Alice'
        >>> with SwapAttr(self, 'person', 'Bob'):
        ...     # ...

    """

    def __init__(self, obj, attr, value):
        # type: (Any, Text, Any) -> None
        self.obj = obj
        self.attr = attr
        self.value = value
        self.old_value = getattr(obj, attr)

    def __enter__(self):
        # type: () -> None
        setattr(self.obj, self.attr, self.value)

    def __exit__(self, *args):
        # type: (Any) -> None
        setattr(self.obj, self.attr, self.old_value)
